from collections.abc import Sequence
from itertools import pairwise
from logging import getLogger
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast, final

from ..errors import WorkflowConstructionError
from ..packet_pool import Packet, PacketPool
from ..processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.io import InT, OutT_co, ProcName
from .workflow_processor import WorkflowProcessor

logger = getLogger(__name__)

_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class ExitWorkflowLoopHandler(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self,
        out_packet: Packet[_OutT_contra],
        ctx: RunContext[CtxT] | None,
        **kwargs: Any,
    ) -> bool: ...


class LoopedWorkflow(
    WorkflowProcessor[InT, OutT_co, CtxT], Generic[InT, OutT_co, CtxT]
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, Any, CtxT]],
        exit_proc: Processor[Any, OutT_co, Any, CtxT],
        packet_pool: PacketPool[CtxT] | None = None,
        recipients: list[ProcName] | None = None,
        num_par_run_retries: int = 0,
        max_iterations: int = 10,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            name=name,
            start_proc=subprocs[0],
            end_proc=exit_proc,
            packet_pool=packet_pool,
            recipients=recipients,
            num_par_run_retries=num_par_run_retries,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor "
                    f"{prev_proc.name} does not match input type {proc.in_type} of "
                    f"subprocessor {proc.name}"
                )
        if subprocs[-1].out_type != subprocs[0].in_type:
            raise WorkflowConstructionError(
                "Looped workflow's last subprocessor output type "
                f"{subprocs[-1].out_type} does not match first subprocessor input "
                f"type {subprocs[0].in_type}"
            )

        self._max_iterations = max_iterations

        self._exit_workflow_loop_impl: ExitWorkflowLoopHandler[OutT_co, CtxT] | None = (
            None
        )

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def exit_workflow_loop(
        self, func: ExitWorkflowLoopHandler[OutT_co, CtxT]
    ) -> ExitWorkflowLoopHandler[OutT_co, CtxT]:
        self._exit_workflow_loop_impl = func

        return func

    def _exit_workflow_loop(
        self,
        out_packet: Packet[OutT_co],
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        if self._exit_workflow_loop_impl:
            return self._exit_workflow_loop_impl(out_packet, ctx=ctx, **kwargs)

        return False

    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        run_id: str | None = None,
        forgetful: bool = False,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        packet = in_packet
        num_iterations = 0
        exit_packet: Packet[OutT_co] | None = None

        while True:
            for subproc in self.subprocs:
                packet = await subproc.run(
                    chat_inputs=chat_inputs,
                    in_packet=packet,
                    in_args=in_args,
                    forgetful=forgetful,
                    run_id=self._generate_subproc_run_id(run_id, subproc=subproc),
                    ctx=ctx,
                )

                if subproc is self._end_proc:
                    num_iterations += 1
                    exit_packet = cast("Packet[OutT_co]", packet)
                    if self._exit_workflow_loop(exit_packet, ctx=ctx):
                        return exit_packet
                    if num_iterations >= self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). "
                            "Exiting loop."
                        )
                        return exit_packet

                chat_inputs = None
                in_args = None



from typing import TypeAlias, TypeVar

from pydantic import BaseModel

ProcName: TypeAlias = str


InT = TypeVar("InT")
OutT_co = TypeVar("OutT_co", covariant=True)


class LLMPromptArgs(BaseModel):
    pass


LLMPrompt: TypeAlias = str



import ast
import asyncio
import json
import re
from collections.abc import Coroutine, Mapping
from datetime import UTC, datetime
from logging import getLogger
from pathlib import Path
from typing import Annotated, Any, TypeVar, get_args, get_origin, overload

from pydantic import TypeAdapter, ValidationError
from tqdm.autonotebook import tqdm

from .errors import OutputValidationError, StringParsingError

logger = getLogger(__name__)

_JSON_START_RE = re.compile(r"[{\[]")

T = TypeVar("T")


def extract_json_substring(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for match in _JSON_START_RE.finditer(text):
        start = match.start()
        try:
            _, end = decoder.raw_decode(text, idx=start)
            return text[start:end]
        except json.JSONDecodeError:
            continue

    return None


def parse_json_or_py_string(
    s: str, return_none_on_failure: bool = False, strip_language_markdown: bool = True
) -> dict[str, Any] | list[Any] | None:
    s_orig = s
    if strip_language_markdown:
        s = re.sub(r"```[a-zA-Z0-9]*\n|```", "", s).strip()
    try:
        return ast.literal_eval(s)
    except (ValueError, SyntaxError):
        try:
            return json.loads(s)
        except json.JSONDecodeError as exc:
            if return_none_on_failure:
                return None
            raise StringParsingError(
                "Invalid JSON/Python string - Both ast.literal_eval and json.loads "
                f"failed to parse the following response:\n{s_orig}"
            ) from exc


def parse_json_or_py_substring(
    json_str: str,
    return_none_on_failure: bool = False,
    strip_language_markdown: bool = True,
) -> dict[str, Any] | list[Any] | None:
    return parse_json_or_py_string(
        extract_json_substring(json_str) or "",
        return_none_on_failure=return_none_on_failure,
        strip_language_markdown=strip_language_markdown,
    )


@overload
def validate_obj_from_json_or_py_string(
    s: str,
    adapter: TypeAdapter[T],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> T: ...


@overload
def validate_obj_from_json_or_py_string(
    s: str,
    adapter: Mapping[str, TypeAdapter[T]],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> T | str: ...


def validate_obj_from_json_or_py_string(
    s: str,
    adapter: TypeAdapter[T] | Mapping[str, TypeAdapter[T]],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> T | str:
    _selected_adapter: TypeAdapter[T] | None = None
    _selected_tag: str | None = None
    s_orig = s

    if isinstance(adapter, Mapping):
        for _tag, _adapter in adapter.items():
            match = re.search(rf"<{_tag}>\s*(.*?)\s*</{_tag}>", s, re.DOTALL)
            if not match:
                continue
            s = match.group(1).strip()
            _selected_adapter = _adapter
            _selected_tag = _tag
            break
        if _selected_adapter is None:
            return s
    else:
        _selected_adapter = adapter

    _type = _selected_adapter._type  # type: ignore[attr-defined]
    type_origin = get_origin(_type)
    type_args = get_args(_type)
    is_str_type = (_type is str) or (
        type_origin is Annotated and type_args and type_args[0] is str
    )

    try:
        if not is_str_type:
            if from_substring:
                parsed = parse_json_or_py_substring(
                    s,
                    return_none_on_failure=True,
                    strip_language_markdown=strip_language_markdown,
                )
            else:
                parsed = parse_json_or_py_string(
                    s,
                    return_none_on_failure=True,
                    strip_language_markdown=strip_language_markdown,
                )
            if parsed is None:
                parsed = s
        else:
            parsed = s
        return _selected_adapter.validate_python(parsed)
    except ValidationError as exc:
        err_message = f"Invalid JSON or Python string:\n{s_orig}"
        if _selected_tag:
            err_message += f"\nExpected type {_type} within tag <{_selected_tag}>"
        else:
            err_message += f"\nExpected type {_type}"
        raise OutputValidationError(err_message) from exc


def extract_xml_list(text: str) -> list[str]:
    pattern = re.compile(r"<(chunk_\d+)>(.*?)</\1>", re.DOTALL)

    chunks: list[str] = []
    for match in pattern.finditer(text):
        content = match.group(2).strip()
        chunks.append(content)
    return chunks


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.exception(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    # TODO: optimize
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now(UTC).strftime("%Y%m%d_%H%M%S")



import asyncio
import json
from collections.abc import AsyncIterator, Coroutine, Sequence
from itertools import starmap
from logging import getLogger
from typing import Any, ClassVar, Generic, Protocol, TypeVar

from pydantic import BaseModel

from .generics_utils import AutoInstanceAttributesMixin
from .llm import LLM, LLMSettings
from .llm_agent_memory import LLMAgentMemory
from .run_context import CtxT, RunContext
from .typing.completion import Completion
from .typing.converters import Converters
from .typing.events import (
    CompletionChunkEvent,
    CompletionEvent,
    Event,
    GenMessageEvent,
    ToolCallEvent,
    ToolMessageEvent,
    UserMessageEvent,
)
from .typing.message import AssistantMessage, Messages, ToolMessage, UserMessage
from .typing.tool import BaseTool, NamedToolChoice, ToolCall, ToolChoice

logger = getLogger(__name__)


FINAL_ANSWER_TOOL_NAME = "final_answer"


_FinalAnswerT = TypeVar("_FinalAnswerT")


class ExitToolCallLoopHandler(Protocol[CtxT]):
    def __call__(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT] | None,
        **kwargs: Any,
    ) -> bool: ...


class ManageMemoryHandler(Protocol[CtxT]):
    def __call__(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT] | None,
        **kwargs: Any,
    ) -> None: ...


class LLMPolicyExecutor(AutoInstanceAttributesMixin, Generic[_FinalAnswerT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_final_answer_type",
    }

    def __init__(
        self,
        agent_name: str,
        llm: LLM[LLMSettings, Converters],
        tools: list[BaseTool[BaseModel, Any, CtxT]] | None,
        max_turns: int,
        react_mode: bool = False,
        final_answer_as_tool_call: bool = False,
    ) -> None:
        self._final_answer_type: type[_FinalAnswerT]
        super().__init__()

        self._agent_name = agent_name

        _tools: list[BaseTool[BaseModel, Any, CtxT]] | None = tools
        self._final_answer_tool_name: str | None = None
        if tools and final_answer_as_tool_call:
            final_answer_tool = self.get_final_answer_tool()
            self._final_answer_tool_name = final_answer_tool.name
            _tools = tools + [final_answer_tool]

        self._llm = llm
        self._llm.tools = _tools

        self._max_turns = max_turns
        self._react_mode = react_mode

        self.exit_tool_call_loop_impl: ExitToolCallLoopHandler[CtxT] | None = None
        self.manage_memory_impl: ManageMemoryHandler[CtxT] | None = None

    @property
    def agent_name(self) -> str:
        return self._agent_name

    @property
    def llm(self) -> LLM[LLMSettings, Converters]:
        return self._llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._llm.tools or {}

    @property
    def max_turns(self) -> int:
        return self._max_turns

    def _exit_tool_call_loop(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        if self.exit_tool_call_loop_impl:
            return self.exit_tool_call_loop_impl(conversation, ctx=ctx, **kwargs)

        assert conversation, "Conversation must not be empty"
        assert isinstance(conversation[-1], AssistantMessage), (
            "Last message in conversation must be an AssistantMessage"
        )

        return not bool(conversation[-1].tool_calls)

    def _manage_memory(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> None:
        if self.manage_memory_impl:
            self.manage_memory_impl(memory=memory, ctx=ctx, **kwargs)

    async def generate_messages(
        self,
        memory: LLMAgentMemory,
        run_id: str,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Sequence[AssistantMessage]:
        completion = await self.llm.generate_completion(
            memory.message_history, tool_choice=tool_choice
        )
        memory.update(completion.messages)

        if ctx is not None:
            ctx.completions[self.agent_name].append(completion)
            self._track_usage(self.agent_name, completion, ctx=ctx)
            self._print_completion(completion, run_id=run_id, ctx=ctx)

        return completion.messages

    async def generate_messages_stream(
        self,
        memory: LLMAgentMemory,
        run_id: str,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[CompletionChunkEvent | CompletionEvent | GenMessageEvent]:
        message_hist = memory.message_history

        completion: Completion | None = None
        async for event in await self.llm.generate_completion_stream(
            message_hist, tool_choice=tool_choice
        ):
            yield event
            if isinstance(event, CompletionEvent):
                completion = event.data
        if completion is None:
            raise RuntimeError("No completion generated during stream.")

        memory.update(completion.messages)

        for message in completion.messages:
            yield GenMessageEvent(name=self.agent_name, data=message)

        if ctx is not None:
            self._track_usage(self.agent_name, completion, ctx=ctx)
            ctx.completions[self.agent_name].append(completion)

    async def call_tools(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> Sequence[ToolMessage]:
        corouts: list[Coroutine[Any, Any, BaseModel]] = []
        for call in calls:
            tool = self.tools[call.tool_name]
            args = json.loads(call.tool_arguments)
            corouts.append(tool(ctx=ctx, **args))

        outs = await asyncio.gather(*corouts)
        tool_messages = list(
            starmap(ToolMessage.from_tool_output, zip(outs, calls, strict=True))
        )
        memory.update(tool_messages)

        if ctx is not None:
            ctx.printer.print_llm_messages(
                tool_messages, agent_name=self.agent_name, run_id=run_id
            )

        return tool_messages

    async def call_tools_stream(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[ToolMessageEvent]:
        tool_messages = await self.call_tools(
            calls, memory=memory, run_id=run_id, ctx=ctx
        )
        for tool_message, call in zip(tool_messages, calls, strict=True):
            yield ToolMessageEvent(name=call.tool_name, data=tool_message)

    def _extract_final_answer_from_tool_calls(
        self, gen_message: AssistantMessage, memory: LLMAgentMemory
    ) -> AssistantMessage | None:
        final_answer_message: AssistantMessage | None = None
        for tool_call in gen_message.tool_calls or []:
            if tool_call.tool_name == self._final_answer_tool_name:
                final_answer_message = AssistantMessage(
                    name=self.agent_name, content=tool_call.tool_arguments
                )
                gen_message.tool_calls = None
                memory.update([final_answer_message])
                return final_answer_message

        return final_answer_message

    async def _generate_final_answer(
        self, memory: LLMAgentMemory, run_id: str, ctx: RunContext[CtxT] | None = None
    ) -> AssistantMessage:
        assert self._final_answer_tool_name is not None

        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!"
        )
        memory.update([user_message])
        if ctx is not None:
            ctx.printer.print_llm_messages(
                [user_message], agent_name=self.agent_name, run_id=run_id
            )

        tool_choice = NamedToolChoice(name=self._final_answer_tool_name)
        gen_message = (
            await self.generate_messages(
                memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
            )
        )[0]

        final_answer_message = self._extract_final_answer_from_tool_calls(
            gen_message, memory=memory
        )
        if final_answer_message is None:
            raise RuntimeError(
                "Final answer tool call did not return a final answer message."
            )

        return final_answer_message

    async def _generate_final_answer_stream(
        self, memory: LLMAgentMemory, run_id: str, ctx: RunContext[CtxT] | None = None
    ) -> AsyncIterator[Event[Any]]:
        assert self._final_answer_tool_name is not None

        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!",
        )
        memory.update([user_message])
        yield UserMessageEvent(name=self.agent_name, data=user_message)

        tool_choice = NamedToolChoice(name=self._final_answer_tool_name)
        event: Event[Any] | None = None
        async for event in self.generate_messages_stream(
            memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
        ):
            yield event

        assert isinstance(event, GenMessageEvent)
        gen_message = event.data
        final_answer_message = self._extract_final_answer_from_tool_calls(
            gen_message, memory=memory
        )
        if final_answer_message is None:
            raise RuntimeError(
                "Final answer tool call did not return a final answer message."
            )
        yield GenMessageEvent(name=self.agent_name, data=final_answer_message)

    async def execute(
        self, memory: LLMAgentMemory, run_id: str, ctx: RunContext[CtxT] | None = None
    ) -> AssistantMessage | Sequence[AssistantMessage]:
        # 1. Generate the first message:
        #    In ReAct mode, we generate the first message without tool calls
        #    to force the agent to plan its actions in a separate message.
        tool_choice: ToolChoice | None = None
        if self.tools:
            tool_choice = "none" if self._react_mode else "auto"
        gen_messages = await self.generate_messages(
            memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
        )
        if not self.tools:
            return gen_messages

        if len(gen_messages) > 1:
            raise ValueError("n_choices must be 1 when executing the tool call loop.")
        gen_message = gen_messages[0]
        turns = 0

        while True:
            # 2. Check if we should exit the tool call loop

            # When final_answer_tool_name is None, we use exit_tool_call_loop_impl
            # to determine whether to exit the loop.
            if self._final_answer_tool_name is None and self._exit_tool_call_loop(
                memory.message_history, ctx=ctx, num_turns=turns
            ):
                return gen_message

            # When final_answer_tool_name is set, we check if the last message contains
            # a tool call to the final answer tool. If it does, we exit the loop.
            if self._final_answer_tool_name is not None:
                final_answer = self._extract_final_answer_from_tool_calls(
                    gen_message, memory=memory
                )
                if final_answer is not None:
                    return final_answer

            # Exit if the maximum number of turns is reached
            if turns >= self.max_turns:
                # When final_answer_tool_name is set, we force the agent to provide
                # a final answer by generating a message with a final answer
                # tool call.
                # Otherwise, we simply return the last generated message.
                if self._final_answer_tool_name is not None:
                    final_answer = await self._generate_final_answer(
                        memory, run_id=run_id, ctx=ctx
                    )
                else:
                    final_answer = gen_message
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return final_answer

            # 3. Call tools if there are any tool calls in the generated message.

            if gen_message.tool_calls:
                await self.call_tools(
                    gen_message.tool_calls, memory=memory, run_id=run_id, ctx=ctx
                )

            # Apply the memory management function if provided.
            self._manage_memory(memory, ctx=ctx, num_turns=turns)

            # 4. Generate the next message based on the updated memory.
            #    In ReAct mode, we set tool_choice to "none" if we just called tools,
            #    so the next message will be an observation/planning message with
            #    no immediate tool calls.
            #    If we are not in ReAct mode, we set tool_choice to "auto" to allow
            #    the LLM to choose freely whether to call tools.

            if self._react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif gen_message.tool_calls:
                tool_choice = "auto"
            else:
                tool_choice = "required"

            gen_message = (
                await self.generate_messages(
                    memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
                )
            )[0]

            turns += 1

    async def execute_stream(
        self, memory: LLMAgentMemory, run_id: str, ctx: RunContext[CtxT] | None = None
    ) -> AsyncIterator[Event[Any]]:
        tool_choice: ToolChoice = "none" if self._react_mode else "auto"
        gen_message: AssistantMessage | None = None
        async for event in self.generate_messages_stream(
            memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
        ):
            yield event
            if isinstance(event, GenMessageEvent):
                gen_message = event.data
        assert isinstance(gen_message, AssistantMessage)

        turns = 0

        while True:
            if self._final_answer_tool_name is None and self._exit_tool_call_loop(
                memory.message_history, ctx=ctx, num_turns=turns
            ):
                return

            if self._final_answer_tool_name is not None:
                final_answer_message = self._extract_final_answer_from_tool_calls(
                    gen_message, memory=memory
                )
                if final_answer_message is not None:
                    yield GenMessageEvent(
                        name=self.agent_name, data=final_answer_message
                    )
                    return

            if turns >= self.max_turns:
                if self._final_answer_tool_name is not None:
                    async for event in self._generate_final_answer_stream(
                        memory, run_id=run_id, ctx=ctx
                    ):
                        yield event
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return

            if gen_message.tool_calls:
                for tool_call in gen_message.tool_calls:
                    yield ToolCallEvent(name=self.agent_name, data=tool_call)

                async for tool_message_event in self.call_tools_stream(
                    gen_message.tool_calls, memory=memory, run_id=run_id, ctx=ctx
                ):
                    yield tool_message_event

            self._manage_memory(memory, ctx=ctx, num_turns=turns)

            if self._react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif gen_message.tool_calls:
                tool_choice = "auto"
            else:
                tool_choice = "required"
            async for event in self.generate_messages_stream(
                memory, tool_choice=tool_choice, run_id=run_id, ctx=ctx
            ):
                yield event
                if isinstance(event, GenMessageEvent):
                    gen_message = event.data

            turns += 1

    def _track_usage(
        self,
        agent_name: str,
        completion: Completion,
        ctx: RunContext[CtxT],
    ) -> None:
        ctx.usage_tracker.update(
            agent_name=agent_name,
            completions=[completion],
            model_name=self.llm.model_name,
        )

    def get_final_answer_tool(self) -> BaseTool[BaseModel, None, Any]:
        if not issubclass(self._final_answer_type, BaseModel):
            raise TypeError(
                "final_answer_type must be a subclass of BaseModel to create "
                "a final answer tool."
            )

        class FinalAnswerTool(BaseTool[self._final_answer_type, None, Any]):
            name: str = FINAL_ANSWER_TOOL_NAME
            description: str = (
                "You must call this tool to provide the final answer. "
                "DO NOT output your answer before calling the tool. "
            )

            async def run(
                self, inp: _FinalAnswerT, ctx: RunContext[Any] | None = None
            ) -> None:
                return None

        return FinalAnswerTool()

    def _print_completion(
        self, completion: Completion, run_id: str, ctx: RunContext[CtxT]
    ) -> None:
        ctx.printer.print_llm_messages(
            completion.messages,
            usages=[completion.usage],
            agent_name=self.agent_name,
            run_id=run_id,
        )



import logging
from collections import defaultdict
from collections.abc import Sequence
from pathlib import Path
from typing import Any, TypeAlias

import yaml
from pydantic import BaseModel, Field
from termcolor import colored

from .typing.completion import Completion, Usage

logger = logging.getLogger(__name__)


COSTS_DICT_PATH: Path = Path(__file__).parent / "costs_dict.yaml"

ModelCostsDict: TypeAlias = dict[str, float]
CostsDict: TypeAlias = dict[str, ModelCostsDict]


class UsageTracker(BaseModel):
    # TODO: specify different costs per provider:model, not just per model
    costs_dict_path: str | Path = COSTS_DICT_PATH
    costs_dict: CostsDict | None = None
    usages: dict[str, Usage] = Field(default_factory=dict)

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.costs_dict = self.load_costs_dict()

    def load_costs_dict(self) -> CostsDict | None:
        try:
            with Path(self.costs_dict_path).open() as f:
                return yaml.safe_load(f)["costs"]
        except Exception:
            logger.info(f"Failed to load cost dictionary from {self.costs_dict_path}")
            return None

    def _add_cost_to_usage(
        self, usage: Usage, model_costs_dict: ModelCostsDict
    ) -> None:
        in_rate = model_costs_dict["input"]
        out_rate = model_costs_dict["output"]
        cached_discount = model_costs_dict.get("cached_discount")
        input_cost = in_rate * usage.input_tokens
        output_cost = out_rate * usage.output_tokens
        reasoning_cost = (
            out_rate * usage.reasoning_tokens
            if usage.reasoning_tokens is not None
            else 0.0
        )
        cached_cost: float = (
            cached_discount * in_rate * usage.cached_tokens
            if (usage.cached_tokens is not None) and (cached_discount is not None)
            else 0.0
        )
        usage.cost = (input_cost + output_cost + reasoning_cost + cached_cost) / 1e6

    def update(
        self,
        agent_name: str,
        completions: Sequence[Completion],
        model_name: str | None = None,
    ) -> None:
        if model_name is not None and self.costs_dict is not None:
            model_costs_dict = self.costs_dict.get(model_name.split(":", 1)[-1])
        else:
            model_costs_dict = None

        for completion in completions:
            if completion.usage is not None:
                if model_costs_dict is not None:
                    self._add_cost_to_usage(
                        usage=completion.usage, model_costs_dict=model_costs_dict
                    )
                if agent_name not in self.usages:
                    self.usages[agent_name] = Usage()
                self.usages[agent_name] += completion.usage

    @property
    def total_usage(self) -> Usage:
        return sum((usage for usage in self.usages.values()), Usage())

    def reset(self) -> None:
        self.usages = defaultdict(Usage)

    def print_usage(self) -> None:
        usage = self.total_usage

        logger.debug("\n-------------------")

        token_usage_str = (
            f"Total I/O/(R)/(C) tokens: {usage.input_tokens}/{usage.output_tokens}"
        )
        if usage.reasoning_tokens is not None:
            token_usage_str += f"/{usage.reasoning_tokens}"
        if usage.cached_tokens is not None:
            token_usage_str += f"/{usage.cached_tokens}"
        logger.debug(colored(token_usage_str, "light_grey"))

        if usage.cost is not None:
            logger.debug(
                colored(
                    f"Total cost: ${usage.cost:.4f}",
                    "light_grey",
                )
            )



from ..errors import CompletionError
from ..typing.completion_chunk import (
    CompletionChunk,
    CompletionChunkChoice,
    CompletionChunkChoiceDelta,
    CompletionChunkDeltaToolCall,
)
from . import OpenAICompletionChunk  # type: ignore[import]
from .completion_converters import from_api_completion_usage


def from_api_completion_chunk(
    api_completion_chunk: OpenAICompletionChunk, name: str | None = None
) -> CompletionChunk:
    if api_completion_chunk.choices is None:  # type: ignore
        raise CompletionError(
            f"Completion chunk API error: "
            f"{getattr(api_completion_chunk, 'error', None)}"
        )

    choices: list[CompletionChunkChoice] = []

    for api_choice in api_completion_chunk.choices:
        index = api_choice.index
        finish_reason = api_choice.finish_reason

        if api_choice.delta is None:  # type: ignore
            raise CompletionError(
                "API returned None for delta content in completion chunk "
                f"with finish_reason: {finish_reason}."
            )
        # if api_choice.delta.content is None:
        #     raise CompletionError(
        #         "API returned None for delta content in completion chunk "
        #         f"with finish_reason: {finish_reason}."
        #     )

        delta = CompletionChunkChoiceDelta(
            content=api_choice.delta.content,
            refusal=api_choice.delta.refusal,
            role=api_choice.delta.role,
            tool_calls=[
                CompletionChunkDeltaToolCall(
                    id=tool_call.id,
                    index=tool_call.index,
                    tool_name=tool_call.function.name,
                    tool_arguments=tool_call.function.arguments,
                )
                for tool_call in (api_choice.delta.tool_calls or [])
                if tool_call.function
            ],
        )

        choice = CompletionChunkChoice(
            index=index,
            delta=delta,
            finish_reason=finish_reason,
            # OpenAI logprobs have identical, but separately defined, types for
            # completion chunks and completions
            logprobs=api_choice.logprobs,
        )

        choices.append(choice)

    usage = (
        from_api_completion_usage(api_completion_chunk.usage)
        if api_completion_chunk.usage
        else None
    )

    return CompletionChunk(
        id=api_completion_chunk.id,
        model=api_completion_chunk.model,
        name=name,
        created=api_completion_chunk.created,
        system_fingerprint=api_completion_chunk.system_fingerprint,
        choices=choices,
        usage=usage,
    )



import fnmatch
import logging
import os
from abc import abstractmethod
from collections.abc import AsyncIterator, Mapping
from copy import deepcopy
from typing import Any, Generic, Literal, NotRequired

import httpx
from pydantic import BaseModel
from tenacity import (
    RetryCallState,
    retry,
    stop_after_attempt,
    wait_random_exponential,
)
from typing_extensions import TypedDict

from .http_client import AsyncHTTPClientParams, create_simple_async_httpx_client
from .llm import LLM, ConvertT_co, LLMSettings, SettingsT_co
from .rate_limiting.rate_limiter_chunked import RateLimiterC, limit_rate
from .typing.completion import Completion
from .typing.completion_chunk import (
    CompletionChoice,
    CompletionChunk,
    combine_completion_chunks,
)
from .typing.events import CompletionChunkEvent, CompletionEvent
from .typing.message import AssistantMessage, Messages
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


APIProviderName = Literal["openai", "openrouter", "google_ai_studio"]


class APIProvider(TypedDict):
    name: APIProviderName
    base_url: str
    api_key: NotRequired[str | None]
    struct_outputs_support: NotRequired[tuple[str, ...]]


API_PROVIDERS: dict[APIProviderName, APIProvider] = {
    "openai": APIProvider(
        name="openai",
        base_url="https://api.openai.com/v1",
        api_key=os.getenv("OPENAI_API_KEY"),
        struct_outputs_support=("*",),
    ),
    "openrouter": APIProvider(
        name="openrouter",
        base_url="https://openrouter.ai/api/v1",
        api_key=os.getenv("OPENROUTER_API_KEY"),
        struct_outputs_support=(),
    ),
    "google_ai_studio": APIProvider(
        name="google_ai_studio",
        base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
        api_key=os.getenv("GOOGLE_AI_STUDIO_API_KEY"),
        struct_outputs_support=("*",),
    ),
}


def retry_error_callback(retry_state: RetryCallState) -> Completion:
    exception = retry_state.outcome.exception() if retry_state.outcome else None
    if exception:
        if retry_state.attempt_number == 1:
            logger.warning(
                f"\nCloudLLM completion request failed:\n{exception}",
                # exc_info=exception,
            )
        if retry_state.attempt_number > 1:
            logger.warning(
                f"\nCloudLLM completion request failed after retrying:\n{exception}",
                # exc_info=exception,
            )
    failed_message = AssistantMessage(content=None, refusal=str(exception))

    return Completion(
        model="",
        choices=[CompletionChoice(message=failed_message, finish_reason=None, index=0)],
    )


def retry_before_sleep_callback(retry_state: RetryCallState) -> None:
    exception = retry_state.outcome.exception() if retry_state.outcome else None
    if exception:
        logger.info(
            "\nRetrying CloudLLM completion request "
            f"(attempt {retry_state.attempt_number}):\n{exception}"
        )


class CloudLLMSettings(LLMSettings, total=False):
    use_struct_outputs: bool


class CloudLLM(LLM[SettingsT_co, ConvertT_co], Generic[SettingsT_co, ConvertT_co]):
    def __init__(
        self,
        # Base LLM args
        model_name: str,
        converters: ConvertT_co,
        llm_settings: SettingsT_co | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | Mapping[str, type] | None = None,
        model_id: str | None = None,
        # Custom LLM provider
        api_provider: APIProvider | None = None,
        # Connection settings
        async_http_client: httpx.AsyncClient | None = None,
        async_http_client_params: (
            dict[str, Any] | AsyncHTTPClientParams | None
        ) = None,
        # Rate limiting
        rate_limiter: (RateLimiterC[Messages, AssistantMessage] | None) = None,
        rate_limiter_rpm: float | None = None,
        rate_limiter_chunk_size: int = 1000,
        rate_limiter_max_concurrency: int = 300,
        # Retries
        num_generation_retries: int = 0,
        **kwargs: Any,
    ) -> None:
        self.llm_settings: CloudLLMSettings | None

        super().__init__(
            model_name=model_name,
            llm_settings=llm_settings,
            converters=converters,
            model_id=model_id,
            tools=tools,
            response_format=response_format,
            **kwargs,
        )

        self._model_name = model_name
        model_name_parts = model_name.split(":", 1)

        if len(model_name_parts) == 2:
            api_provider_name, api_model_name = model_name_parts
            self._api_model_name: str = api_model_name
            if api_provider_name not in API_PROVIDERS:
                raise ValueError(
                    f"API provider '{api_provider_name}' is not supported. "
                    f"Supported providers are: {', '.join(API_PROVIDERS.keys())}"
                )
            _api_provider = API_PROVIDERS[api_provider_name]
        elif api_provider is not None:
            self._api_model_name: str = model_name
            _api_provider = api_provider
        else:
            raise ValueError(
                "API provider must be specified either in the model name "
                "or as a separate argument."
            )

        self._api_provider_name: APIProviderName = _api_provider["name"]
        self._base_url: str | None = _api_provider.get("base_url")
        self._api_key: str | None = _api_provider.get("api_key")
        self._struct_outputs_support: bool = any(
            fnmatch.fnmatch(self._model_name, pat)
            for pat in _api_provider.get("struct_outputs_support", ())
        )

        if (
            self._llm_settings.get("use_struct_outputs")
            and not self._struct_outputs_support
        ):
            raise ValueError(
                f"Model {self._model_name} does not support structured outputs."
            )

        self._tool_call_settings: dict[str, Any] = {}

        self._rate_limiter: RateLimiterC[Messages, AssistantMessage] | None = (
            self._get_rate_limiter(
                rate_limiter=rate_limiter,
                rpm=rate_limiter_rpm,
                chunk_size=rate_limiter_chunk_size,
                max_concurrency=rate_limiter_max_concurrency,
            )
        )

        self._async_http_client: httpx.AsyncClient | None = None
        if async_http_client is not None:
            self._async_http_client = async_http_client
        elif async_http_client_params is not None:
            self._async_http_client = create_simple_async_httpx_client(
                async_http_client_params
            )

        self.num_generation_retries = num_generation_retries

    @property
    def api_provider_name(self) -> APIProviderName | None:
        return self._api_provider_name

    @property
    def rate_limiter(
        self,
    ) -> RateLimiterC[Messages, AssistantMessage] | None:
        return self._rate_limiter

    def _make_completion_kwargs(
        self,
        conversation: Messages,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> dict[str, Any]:
        api_messages = [self._converters.to_message(m) for m in conversation]

        api_tools = None
        api_tool_choice = None
        if self.tools:
            api_tools = [
                self._converters.to_tool(t, **self._tool_call_settings)
                for t in self.tools.values()
            ]
            if tool_choice is not None:
                api_tool_choice = self._converters.to_tool_choice(tool_choice)

        api_llm_settings = deepcopy(self.llm_settings or {})
        api_llm_settings.pop("use_struct_outputs", None)

        return dict(
            api_messages=api_messages,
            api_tools=api_tools,
            api_tool_choice=api_tool_choice,
            api_response_format=self._response_format,
            n_choices=n_choices,
            **api_llm_settings,
        )

    @abstractmethod
    async def _get_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_parsed_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_format: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_completion_stream(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[Any]:
        pass

    @abstractmethod
    async def _get_parsed_completion_stream(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_format: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[Any]:
        pass

    async def generate_completion_no_retry(
        self,
        conversation: Messages,
        *,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> Completion:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation, tool_choice=tool_choice, n_choices=n_choices
        )

        if not self._llm_settings.get("use_struct_outputs"):
            completion_kwargs.pop("api_response_format", None)
            api_completion = await self._get_completion(**completion_kwargs)
        else:
            api_completion = await self._get_parsed_completion(**completion_kwargs)

        completion = self._converters.from_completion(
            api_completion, name=self.model_id
        )

        if not self._llm_settings.get("use_struct_outputs"):
            # If validation is not handled by the structured output functionality
            # of the LLM provider
            self._validate_completion(completion)
            self._validate_tool_calls(completion)

        return completion

    async def generate_completion_stream(
        self,
        conversation: Messages,
        *,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> AsyncIterator[CompletionChunkEvent | CompletionEvent]:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation, tool_choice=tool_choice, n_choices=n_choices
        )

        if not self._llm_settings.get("use_struct_outputs"):
            completion_kwargs.pop("api_response_format", None)
            api_stream = await self._get_completion_stream(**completion_kwargs)
        else:
            api_stream = await self._get_parsed_completion_stream(**completion_kwargs)

        async def iterate() -> AsyncIterator[CompletionChunkEvent | CompletionEvent]:
            completion_chunks: list[CompletionChunk] = []
            async for api_completion_chunk in api_stream:
                completion_chunk = self._converters.from_completion_chunk(
                    api_completion_chunk, name=self.model_id
                )
                completion_chunks.append(completion_chunk)
                yield CompletionChunkEvent(data=completion_chunk, name=self.model_id)

            # TODO: can be done using the OpenAI final_completion_chunk
            completion = combine_completion_chunks(completion_chunks)

            yield CompletionEvent(data=completion, name=self.model_id)

            if not self._llm_settings.get("use_struct_outputs"):
                # If validation is not handled by the structured outputs functionality
                # of the LLM provider
                self._validate_completion(completion)
                self._validate_tool_calls(completion)

        return iterate()

    @limit_rate
    async def generate_completion(  # type: ignore[override]
        self,
        conversation: Messages,
        *,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> Completion:
        wrapped_func = retry(
            wait=wait_random_exponential(min=1, max=8),
            stop=stop_after_attempt(self.num_generation_retries + 1),
            before_sleep=retry_before_sleep_callback,
            retry_error_callback=retry_error_callback,
        )(self.__class__.generate_completion_no_retry)

        return await wrapped_func(
            self, conversation, tool_choice=tool_choice, n_choices=n_choices
        )

    def _get_rate_limiter(
        self,
        rate_limiter: RateLimiterC[Messages, AssistantMessage] | None = None,
        rpm: float | None = None,
        chunk_size: int = 1000,
        max_concurrency: int = 300,
    ) -> RateLimiterC[Messages, AssistantMessage] | None:
        if rate_limiter is not None:
            logger.info(
                f"[{self.__class__.__name__}] Set rate limit to {rate_limiter.rpm} RPM"
            )
            return rate_limiter
        if rpm is not None:
            logger.info(f"[{self.__class__.__name__}] Set rate limit to {rpm} RPM")
            return RateLimiterC(
                rpm=rpm, chunk_size=chunk_size, max_concurrency=max_concurrency
            )

        return None



import logging
from collections.abc import AsyncIterator, Iterable, Mapping
from copy import deepcopy
from typing import Any, Literal, NamedTuple

import httpx
from openai import AsyncOpenAI
from openai._types import NOT_GIVEN  # type: ignore[import]
from openai.lib.streaming.chat import (
    AsyncChatCompletionStreamManager as OpenAIAsyncChatCompletionStreamManager,
)
from openai.lib.streaming.chat import ChunkEvent as OpenAIChunkEvent
from pydantic import BaseModel

from ..cloud_llm import APIProvider, CloudLLM, CloudLLMSettings
from ..http_client import AsyncHTTPClientParams
from ..rate_limiting.rate_limiter_chunked import RateLimiterC
from ..typing.message import AssistantMessage, Messages
from ..typing.tool import BaseTool
from . import (
    OpenAICompletion,
    OpenAICompletionChunk,
    OpenAIMessageParam,
    OpenAIParsedCompletion,
    OpenAIPredictionContentParam,
    OpenAIStreamOptionsParam,
    OpenAIToolChoiceOptionParam,
    OpenAIToolParam,
)
from .converters import OpenAIConverters

logger = logging.getLogger(__name__)


class ToolCallSettings(NamedTuple):
    strict: bool | None = None


class OpenAILLMSettings(CloudLLMSettings, total=False):
    reasoning_effort: Literal["low", "medium", "high"] | None

    parallel_tool_calls: bool

    modalities: list[Literal["text", "audio"]] | None

    frequency_penalty: float | None
    presence_penalty: float | None
    logit_bias: dict[str, int] | None
    stop: str | list[str] | None
    logprobs: bool | None
    top_logprobs: int | None

    prediction: OpenAIPredictionContentParam | None

    stream_options: OpenAIStreamOptionsParam | None

    metadata: dict[str, str] | None
    store: bool | None
    user: str

    # response_format: (
    #     OpenAIResponseFormatText
    #     | OpenAIResponseFormatJSONSchema
    #     | OpenAIResponseFormatJSONObject
    # )


class OpenAILLM(CloudLLM[OpenAILLMSettings, OpenAIConverters]):
    def __init__(
        self,
        # Base LLM args
        model_name: str,
        llm_settings: OpenAILLMSettings | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | Mapping[str, type] | None = None,
        model_id: str | None = None,
        # Custom LLM provider
        api_provider: APIProvider | None = None,
        # Connection settings
        async_http_client: httpx.AsyncClient | None = None,
        async_http_client_params: (
            dict[str, Any] | AsyncHTTPClientParams | None
        ) = None,
        async_openai_client_params: dict[str, Any] | None = None,
        # Rate limiting
        rate_limiter: (RateLimiterC[Messages, AssistantMessage] | None) = None,
        rate_limiter_rpm: float | None = None,
        rate_limiter_chunk_size: int = 1000,
        rate_limiter_max_concurrency: int = 300,
        # Retries
        num_generation_retries: int = 0,
    ) -> None:
        super().__init__(
            model_name=model_name,
            model_id=model_id,
            llm_settings=llm_settings,
            converters=OpenAIConverters(),
            tools=tools,
            response_format=response_format,
            api_provider=api_provider,
            async_http_client=async_http_client,
            async_http_client_params=async_http_client_params,
            rate_limiter=rate_limiter,
            rate_limiter_rpm=rate_limiter_rpm,
            rate_limiter_chunk_size=rate_limiter_chunk_size,
            rate_limiter_max_concurrency=rate_limiter_max_concurrency,
            num_generation_retries=num_generation_retries,
        )

        self._tool_call_settings = {
            "strict": self._llm_settings.get("use_struct_outputs", False),
        }

        _async_openai_client_params = deepcopy(async_openai_client_params or {})
        if self._async_http_client is not None:
            _async_openai_client_params["http_client"] = self._async_http_client

        self._client: AsyncOpenAI = AsyncOpenAI(
            base_url=self._base_url,
            api_key=self._api_key,
            **_async_openai_client_params,
        )

    async def _get_completion(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> OpenAICompletion:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        n = n_choices or NOT_GIVEN

        return await self._client.chat.completions.create(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            n=n,
            stream=False,
            **api_llm_settings,
        )

    async def _get_parsed_completion(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_format: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> OpenAIParsedCompletion[Any]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        n = n_choices or NOT_GIVEN
        response_format = api_response_format or NOT_GIVEN

        return await self._client.beta.chat.completions.parse(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            response_format=response_format,
            n=n,
            **api_llm_settings,
        )

    async def _get_completion_stream(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[OpenAICompletionChunk]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        n = n_choices or NOT_GIVEN

        stream_generator = await self._client.chat.completions.create(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            stream=True,
            n=n,
            **api_llm_settings,
        )

        async def iterate() -> AsyncIterator[OpenAICompletionChunk]:
            async with stream_generator as stream:
                async for completion_chunk in stream:
                    yield completion_chunk

        return iterate()

    async def _get_parsed_completion_stream(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_format: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[OpenAICompletionChunk]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        response_format = api_response_format or NOT_GIVEN
        n = n_choices or NOT_GIVEN

        stream_manager: OpenAIAsyncChatCompletionStreamManager[
            OpenAICompletionChunk
        ] = self._client.beta.chat.completions.stream(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            response_format=response_format,
            n=n,
            **api_llm_settings,
        )

        async def iterate() -> AsyncIterator[OpenAICompletionChunk]:
            async with stream_manager as stream:
                async for chunk_event in stream:
                    if isinstance(chunk_event, OpenAIChunkEvent):
                        yield chunk_event.chunk

        return iterate()



import json
from collections.abc import Mapping, Sequence
from typing import ClassVar, Generic, Protocol, TypeAlias, TypeVar

from pydantic import BaseModel, TypeAdapter

from .errors import InputPromptBuilderError, SystemPromptBuilderError
from .generics_utils import AutoInstanceAttributesMixin
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.io import InT, LLMPrompt, LLMPromptArgs
from .typing.message import UserMessage

_InT_contra = TypeVar("_InT_contra", contravariant=True)


class MakeSystemPromptHandler(Protocol[CtxT]):
    def __call__(
        self,
        sys_args: LLMPromptArgs | None,
        *,
        ctx: RunContext[CtxT] | None,
    ) -> str: ...


class MakeInputContentHandler(Protocol[_InT_contra, CtxT]):
    def __call__(
        self,
        *,
        in_args: _InT_contra | None,
        usr_args: LLMPromptArgs | None,
        ctx: RunContext[CtxT] | None,
    ) -> Content: ...


PromptArgumentType: TypeAlias = str | bool | int | ImageData


class PromptBuilder(AutoInstanceAttributesMixin, Generic[InT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {0: "_in_type"}

    def __init__(
        self,
        agent_name: str,
        sys_prompt_template: LLMPrompt | None,
        in_prompt_template: LLMPrompt | None,
        sys_args_schema: type[LLMPromptArgs] | None = None,
        usr_args_schema: type[LLMPromptArgs] | None = None,
    ):
        self._in_type: type[InT]
        super().__init__()

        self._agent_name = agent_name
        self.sys_prompt_template = sys_prompt_template
        self.in_prompt_template = in_prompt_template
        self.sys_args_schema = sys_args_schema
        self.usr_args_schema = usr_args_schema
        self.make_system_prompt_impl: MakeSystemPromptHandler[CtxT] | None = None
        self.make_input_content_impl: MakeInputContentHandler[InT, CtxT] | None = None

        self._in_args_type_adapter: TypeAdapter[InT] = TypeAdapter(self._in_type)

    def make_system_prompt(
        self, sys_args: LLMPromptArgs | None = None, ctx: RunContext[CtxT] | None = None
    ) -> str | None:
        if self.sys_prompt_template is None:
            return None

        val_sys_args = sys_args
        if sys_args is not None:
            if self.sys_args_schema is not None:
                val_sys_args = self.sys_args_schema.model_validate(sys_args)
            else:
                raise SystemPromptBuilderError(
                    "System prompt template and arguments is set, but system arguments "
                    "schema is not provided."
                )

        if self.make_system_prompt_impl:
            return self.make_system_prompt_impl(sys_args=val_sys_args, ctx=ctx)

        sys_args_dict = val_sys_args.model_dump() if val_sys_args else {}

        return self.sys_prompt_template.format(**sys_args_dict)

    def make_input_content(
        self,
        *,
        in_args: InT | None,
        usr_args: LLMPromptArgs | None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Content:
        val_in_args, val_usr_args = self._validate_prompt_args(
            in_args=in_args, usr_args=usr_args
        )

        if self.make_input_content_impl:
            return self.make_input_content_impl(
                in_args=val_in_args, usr_args=val_usr_args, ctx=ctx
            )

        combined_args = self._combine_args(in_args=val_in_args, usr_args=val_usr_args)
        if isinstance(combined_args, str):
            return Content.from_text(combined_args)

        if self.in_prompt_template is not None:
            return Content.from_formatted_prompt(
                self.in_prompt_template, **combined_args
            )

        return Content.from_text(json.dumps(combined_args, indent=2))

    def make_user_message(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        in_args: InT | None = None,
        usr_args: LLMPromptArgs | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> UserMessage | None:
        if chat_inputs is None and in_args is None and usr_args is None:
            return None

        if chat_inputs:
            if isinstance(chat_inputs, LLMPrompt):
                return UserMessage.from_text(chat_inputs, name=self._agent_name)
            return UserMessage.from_content_parts(chat_inputs, name=self._agent_name)

        in_content = self.make_input_content(
            in_args=in_args, usr_args=usr_args, ctx=ctx
        )

        return UserMessage(content=in_content, name=self._agent_name)

    def _validate_prompt_args(
        self,
        *,
        in_args: InT | None,
        usr_args: LLMPromptArgs | None,
    ) -> tuple[InT | None, LLMPromptArgs | None]:
        val_usr_args = usr_args
        if usr_args is not None:
            if self.in_prompt_template is None:
                raise InputPromptBuilderError(
                    "Input prompt template is not set, but user arguments are provided."
                )
            if self.usr_args_schema is None:
                raise InputPromptBuilderError(
                    "User arguments schema is not provided, but user arguments are "
                    "given."
                )
            val_usr_args = self.usr_args_schema.model_validate(usr_args)

        val_in_args = in_args
        if in_args is not None:
            val_in_args = self._in_args_type_adapter.validate_python(in_args)
            if isinstance(val_in_args, BaseModel):
                has_image = self._has_image_data(val_in_args)
                if has_image and self.in_prompt_template is None:
                    raise InputPromptBuilderError(
                        "BaseModel input arguments contain ImageData, but input prompt "
                        "template is not set. Cannot format input arguments."
                    )
            elif self.in_prompt_template is not None:
                raise InputPromptBuilderError(
                    "Cannot use the input prompt template with "
                    "non-BaseModel input arguments."
                )

        return val_in_args, val_usr_args

    @staticmethod
    def _has_image_data(inp: BaseModel) -> bool:
        contains_image_data = False
        for field in type(inp).model_fields:
            if isinstance(getattr(inp, field), ImageData):
                contains_image_data = True

        return contains_image_data

    @staticmethod
    def _format_pydantic_prompt_args(inp: BaseModel) -> dict[str, PromptArgumentType]:
        formatted_args: dict[str, PromptArgumentType] = {}
        for field in type(inp).model_fields:
            if field == "selected_recipients":
                continue

            val = getattr(inp, field)
            if isinstance(val, (int, str, bool, ImageData)):
                formatted_args[field] = val
            else:
                formatted_args[field] = (
                    TypeAdapter(type(val))  # type: ignore[return-value]
                    .dump_json(val, indent=2, warnings="error")
                    .decode("utf-8")
                )

        return formatted_args

    def _combine_args(
        self, *, in_args: InT | None, usr_args: LLMPromptArgs | None
    ) -> Mapping[str, PromptArgumentType] | str:
        fmt_usr_args = self._format_pydantic_prompt_args(usr_args) if usr_args else {}

        if in_args is None:
            return fmt_usr_args

        if isinstance(in_args, BaseModel):
            fmt_in_args = self._format_pydantic_prompt_args(in_args)
            return fmt_in_args | fmt_usr_args

        combined_args_str = self._in_args_type_adapter.dump_json(
            in_args, indent=2, warnings="error"
        ).decode("utf-8")
        if usr_args is not None:
            fmt_usr_args_str = usr_args.model_dump_json(indent=2, warnings="error")
            combined_args_str += "\n" + fmt_usr_args_str

        return combined_args_str



from collections.abc import Sequence
from typing import Generic, TypeVar
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

from .typing.io import ProcName

_PayloadT_co = TypeVar("_PayloadT_co", covariant=True)


class Packet(BaseModel, Generic[_PayloadT_co]):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    payloads: Sequence[_PayloadT_co]
    sender: ProcName
    recipients: Sequence[ProcName] = Field(default_factory=list)

    model_config = ConfigDict(extra="forbid", frozen=True)

    def __repr__(self) -> str:
        return (
            f"From: {self.sender}, To: {', '.join(self.recipients)}, "
            f"Payloads: {len(self.payloads)}"
        )



import json
from collections.abc import Hashable, Mapping, Sequence
from enum import StrEnum
from typing import Annotated, Any, Literal, TypeAlias
from uuid import uuid4

from pydantic import BaseModel, Field
from pydantic.json import pydantic_encoder

from .content import Content, ImageData
from .tool import ToolCall


class Role(StrEnum):
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"
    TOOL = "tool"


class MessageBase(BaseModel):
    id: Hashable = Field(default_factory=lambda: str(uuid4())[:8])
    name: str | None = None


class AssistantMessage(MessageBase):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT
    content: str | None
    tool_calls: Sequence[ToolCall] | None = None
    refusal: str | None = None


class UserMessage(MessageBase):
    role: Literal[Role.USER] = Role.USER
    content: Content | str

    @classmethod
    def from_text(cls, text: str, name: str | None = None) -> "UserMessage":
        return cls(content=Content.from_text(text), name=name)

    @classmethod
    def from_formatted_prompt(
        cls,
        prompt_template: str,
        name: str | None = None,
        prompt_args: Mapping[str, str | int | bool | ImageData] | None = None,
    ) -> "UserMessage":
        content = Content.from_formatted_prompt(prompt_template, **(prompt_args or {}))

        return cls(content=content, name=name)

    @classmethod
    def from_content_parts(
        cls,
        content_parts: Sequence[str | ImageData],
        name: str | None = None,
    ) -> "UserMessage":
        content = Content.from_content_parts(content_parts)

        return cls(content=content, name=name)


class SystemMessage(MessageBase):
    role: Literal[Role.SYSTEM] = Role.SYSTEM
    content: str


class ToolMessage(MessageBase):
    role: Literal[Role.TOOL] = Role.TOOL
    content: str
    tool_call_id: str

    @classmethod
    def from_tool_output(
        cls, tool_output: Any, tool_call: ToolCall, indent: int = 2
    ) -> "ToolMessage":
        return cls(
            content=json.dumps(tool_output, default=pydantic_encoder, indent=indent),
            tool_call_id=tool_call.id,
            name=tool_call.tool_name,
        )


Message = Annotated[
    AssistantMessage | UserMessage | SystemMessage | ToolMessage,
    Field(discriminator="role"),
]

Messages: TypeAlias = list[Message]



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Mapping
from typing import Any, Generic, TypeVar, cast
from uuid import uuid4

from pydantic import BaseModel, TypeAdapter
from typing_extensions import TypedDict

from grasp_agents.utils import validate_obj_from_json_or_py_string

from .errors import ToolValidationError
from .typing.completion import Completion
from .typing.converters import Converters
from .typing.events import CompletionChunkEvent, CompletionEvent
from .typing.message import Messages
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


class LLMSettings(TypedDict, total=False):
    max_completion_tokens: int | None
    temperature: float | None
    top_p: float | None
    seed: int | None


SettingsT_co = TypeVar("SettingsT_co", bound=LLMSettings, covariant=True)
ConvertT_co = TypeVar("ConvertT_co", bound=Converters, covariant=True)


class LLM(ABC, Generic[SettingsT_co, ConvertT_co]):
    @abstractmethod
    def __init__(
        self,
        converters: ConvertT_co,
        model_name: str | None = None,
        model_id: str | None = None,
        llm_settings: SettingsT_co | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: Any | Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()

        self._converters = converters
        self._model_id = model_id or str(uuid4())[:8]
        self._model_name = model_name
        self._tools = {t.name: t for t in tools} if tools else None
        self._llm_settings: SettingsT_co = llm_settings or cast("SettingsT_co", {})

        self._response_format = response_format
        self._response_format_adapter: (
            TypeAdapter[Any] | Mapping[str, TypeAdapter[Any]]
        ) = self._get_response_format_adapter(response_format=response_format)

    @staticmethod
    def _get_response_format_adapter(
        response_format: Any | Mapping[str, Any] | None = None,
    ) -> TypeAdapter[Any] | Mapping[str, TypeAdapter[Any]]:
        if response_format is None:
            return TypeAdapter(Any)
        if isinstance(response_format, Mapping):
            return {k: TypeAdapter(v) for k, v in response_format.items()}  # type: ignore[return-value]
        return TypeAdapter(response_format)

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def model_name(self) -> str | None:
        return self._model_name

    @property
    def llm_settings(self) -> SettingsT_co:
        return self._llm_settings

    @property
    def response_format(self) -> Any | Mapping[str, Any] | None:
        return self._response_format

    @response_format.setter
    def response_format(self, response_format: Any | Mapping[str, Any] | None) -> None:
        self._response_format = response_format
        self._response_format_adapter = self._get_response_format_adapter(
            response_format
        )

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, Any]] | None:
        return self._tools

    @tools.setter
    def tools(self, tools: list[BaseTool[BaseModel, Any, Any]] | None) -> None:
        self._tools = {t.name: t for t in tools} if tools else None

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(model_id={self.model_id}; "
            f"model_name={self._model_name})"
        )

    def _validate_completion(self, completion: Completion) -> None:
        for message in completion.messages:
            if not message.tool_calls:
                validate_obj_from_json_or_py_string(
                    message.content or "",
                    adapter=self._response_format_adapter,
                    from_substring=True,
                )

    def _validate_tool_calls(self, completion: Completion) -> None:
        for message in completion.messages:
            if message.tool_calls:
                for tool_call in message.tool_calls:
                    tool_name = tool_call.tool_name
                    tool_arguments = tool_call.tool_arguments

                    available_tool_names = list(self.tools) if self.tools else []
                    if tool_name not in available_tool_names or not self.tools:
                        raise ToolValidationError(
                            f"Tool '{tool_name}' is not available in the LLM tools "
                            f"(available: {available_tool_names}"
                        )
                    tool = self.tools[tool_name]
                    validate_obj_from_json_or_py_string(
                        tool_arguments, adapter=TypeAdapter(tool.in_type)
                    )

    @abstractmethod
    async def generate_completion(
        self,
        conversation: Messages,
        *,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> Completion:
        pass

    @abstractmethod
    async def generate_completion_stream(
        self,
        conversation: Messages,
        *,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> AsyncIterator[CompletionChunkEvent | CompletionEvent]:
        pass



import time
from collections import defaultdict
from collections.abc import Sequence
from uuid import uuid4

from openai.types.chat.chat_completion_chunk import (
    ChoiceLogprobs as CompletionChunkChoiceLogprobs,
)
from openai.types.chat.chat_completion_token_logprob import (
    ChatCompletionTokenLogprob as CompletionTokenLogprob,
)
from pydantic import BaseModel, Field

from ..errors import CombineCompletionChunksError
from .completion import (
    Completion,
    CompletionChoice,
    CompletionChoiceLogprobs,
    FinishReason,
    Usage,
)
from .message import AssistantMessage, ToolCall


class CompletionChunkDeltaToolCall(BaseModel):
    id: str | None
    index: int
    tool_name: str | None
    tool_arguments: str | None


class CompletionChunkChoiceDelta(BaseModel):
    content: str | None = None
    refusal: str | None
    role: str | None
    tool_calls: list[CompletionChunkDeltaToolCall] | None


class CompletionChunkChoice(BaseModel):
    delta: CompletionChunkChoiceDelta
    finish_reason: FinishReason | None
    index: int
    logprobs: CompletionChunkChoiceLogprobs | None = None


class CompletionChunk(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    name: str | None = None
    system_fingerprint: str | None = None
    choices: list[CompletionChunkChoice]
    usage: Usage | None = None


def combine_completion_chunks(chunks: list[CompletionChunk]) -> Completion:
    if not chunks:
        raise CombineCompletionChunksError(
            "Cannot combine an empty list of completion chunks."
        )

    model_list = {chunk.model for chunk in chunks}
    if len(model_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same model.")
    model = model_list.pop()

    name_list = {chunk.name for chunk in chunks}
    if len(name_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same name.")
    name = name_list.pop()

    system_fingerprints_list = {chunk.system_fingerprint for chunk in chunks}
    if len(system_fingerprints_list) > 1:
        raise CombineCompletionChunksError(
            "All chunks must have the same system fingerprint."
        )
    system_fingerprint = system_fingerprints_list.pop()

    created_list = [chunk.created for chunk in chunks]
    created = max(created_list)

    # Usage is found in the last completion chunk if requested
    usage = chunks[-1].usage

    logp_contents_per_choice: defaultdict[int, list[CompletionTokenLogprob]] = (
        defaultdict(list)
    )
    logp_refusals_per_choice: defaultdict[int, list[CompletionTokenLogprob]] = (
        defaultdict(list)
    )
    logprobs_per_choice: defaultdict[int, CompletionChoiceLogprobs | None] = (
        defaultdict(lambda: None)
    )

    finish_reasons_per_choice: defaultdict[int, FinishReason | None] = defaultdict(
        lambda: None
    )

    contents_per_choice: defaultdict[int, str] = defaultdict(lambda: "")
    refusals_per_choice: defaultdict[int, str] = defaultdict(lambda: "")

    tool_calls_per_choice: defaultdict[
        int, Sequence[CompletionChunkDeltaToolCall] | None
    ] = defaultdict(lambda: None)

    messages_per_choice: dict[int, AssistantMessage] = {}

    for chunk in chunks:
        for choice in chunk.choices:
            index = choice.index

            # Concatenate content and refusal tokens for each choice
            contents_per_choice[index] += choice.delta.content or ""
            refusals_per_choice[index] += choice.delta.refusal or ""

            # Concatenate logprobs for content and refusal tokens for each choice
            if choice.logprobs is not None:
                logp_contents_per_choice[index].extend(choice.logprobs.content or [])
                logp_refusals_per_choice[index].extend(choice.logprobs.refusal or [])

            # Take the last finish reason for each choice
            finish_reasons_per_choice[index] = choice.finish_reason

            # Tool calls should be in the last chunk for each choice
            tool_calls_per_choice[index] = choice.delta.tool_calls

    for index in finish_reasons_per_choice:
        tool_calls: list[ToolCall] = []
        if tool_calls_per_choice[index] is not None:
            for _tool_call in tool_calls_per_choice[index]:  # type: ignore
                if (
                    _tool_call.id is None
                    or _tool_call.tool_name is None
                    or _tool_call.tool_arguments is None
                ):
                    raise CombineCompletionChunksError(
                        "Completion chunk tool calls must have id, tool_name, "
                        "and tool_arguments set."
                    )
                tool_calls.append(
                    ToolCall(
                        id=_tool_call.id,
                        tool_name=_tool_call.tool_name,
                        tool_arguments=_tool_call.tool_arguments,
                    )
                )

        messages_per_choice[index] = AssistantMessage(
            name=name,
            content=contents_per_choice[index] or "<empty>",
            refusal=(refusals_per_choice[index] or None),
            tool_calls=(tool_calls or None),
        )

        if logp_contents_per_choice[index] or logp_refusals_per_choice[index]:
            logprobs_per_choice[index] = CompletionChoiceLogprobs(
                content=logp_contents_per_choice[index],
                refusal=logp_refusals_per_choice[index],
            )

    choices = [
        CompletionChoice(
            index=index,
            message=message,
            finish_reason=finish_reasons_per_choice[index],
            logprobs=logprobs_per_choice[index],
        )
        for index, message in messages_per_choice.items()
    ]

    return Completion(
        model=model,
        name=name,
        created=created,
        system_fingerprint=system_fingerprint,
        choices=choices,
        usage=usage,
    )



import logging
from collections.abc import AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast

from pydantic import BaseModel
from pydantic.json_schema import SkipJsonSchema

from .memory import MemT
from .packet import Packet
from .packet_pool import PacketPool
from .processor import Processor
from .run_context import CtxT, RunContext
from .typing.events import Event, PacketEvent
from .typing.io import InT, OutT_co, ProcName

logger = logging.getLogger(__name__)


class DynCommPayload(BaseModel):
    selected_recipients: SkipJsonSchema[Sequence[ProcName]]


_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class ExitCommunicationHandler(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self,
        out_packet: Packet[_OutT_contra],
        ctx: RunContext[CtxT] | None,
    ) -> bool: ...


class CommProcessor(
    Processor[InT, OutT_co, MemT, CtxT],
    Generic[InT, OutT_co, MemT, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        *,
        recipients: Sequence[ProcName] | None = None,
        packet_pool: PacketPool[CtxT] | None = None,
        num_par_run_retries: int = 0,
    ) -> None:
        super().__init__(name=name, num_par_run_retries=num_par_run_retries)

        self.recipients = recipients or []

        self._packet_pool = packet_pool
        self._is_listening = False
        self._exit_communication_impl: (
            ExitCommunicationHandler[OutT_co, CtxT] | None
        ) = None

    @property
    def packet_pool(self) -> PacketPool[CtxT] | None:
        return self._packet_pool

    def _validate_routing(self, payloads: Sequence[OutT_co]) -> Sequence[ProcName]:
        if all(isinstance(p, DynCommPayload) for p in payloads):
            payloads_ = cast("Sequence[DynCommPayload]", payloads)
            selected_recipients_per_payload = [
                set(p.selected_recipients or []) for p in payloads_
            ]
            assert all(
                x == selected_recipients_per_payload[0]
                for x in selected_recipients_per_payload
            ), "All payloads must have the same recipient IDs for dynamic routing"

            assert payloads_[0].selected_recipients is not None
            selected_recipients = payloads_[0].selected_recipients

            assert all(rid in self.recipients for rid in selected_recipients), (
                "Dynamic routing is enabled, but recipient IDs are not in "
                "the allowed agent's recipient IDs"
            )

            return selected_recipients

        if all((not isinstance(p, DynCommPayload)) for p in payloads):
            return self.recipients

        raise ValueError(
            "All payloads must be either DCommAgentPayload or not DCommAgentPayload"
        )

    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        out_packet = await super().run(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            forgetful=forgetful,
            run_id=run_id,
            ctx=ctx,
        )
        recipients = self._validate_routing(out_packet.payloads)
        routed_out_packet = Packet(
            payloads=out_packet.payloads, sender=self.name, recipients=recipients
        )
        if self._packet_pool is not None and in_packet is None and in_args is None:
            # If no input packet or args, we assume this is the first run.
            await self._packet_pool.post(routed_out_packet)

        return routed_out_packet

    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        out_packet: Packet[OutT_co] | None = None
        async for event in super().run_stream(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            forgetful=forgetful,
            run_id=run_id,
            ctx=ctx,
        ):
            if isinstance(event, PacketEvent):
                out_packet = event.data
            else:
                yield event

        if out_packet is None:
            raise RuntimeError("No output packet generated during stream run")

        recipients = self._validate_routing(out_packet.payloads)
        routed_out_packet = Packet(
            payloads=out_packet.payloads, sender=self.name, recipients=recipients
        )
        if self._packet_pool is not None and in_packet is None and in_args is None:
            # If no input packet or args, we assume this is the first run.
            await self._packet_pool.post(routed_out_packet)

        yield PacketEvent(data=routed_out_packet, name=self.name)

    def exit_communication(
        self, func: ExitCommunicationHandler[OutT_co, CtxT]
    ) -> ExitCommunicationHandler[OutT_co, CtxT]:
        self._exit_communication_impl = func

        return func

    def _exit_communication(
        self, out_packet: Packet[OutT_co], ctx: RunContext[CtxT] | None
    ) -> bool:
        if self._exit_communication_impl:
            return self._exit_communication_impl(out_packet=out_packet, ctx=ctx)

        return False

    async def _packet_handler(
        self,
        packet: Packet[InT],
        ctx: RunContext[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        assert self._packet_pool is not None, "Packet pool must be initialized"

        out_packet = await self.run(ctx=ctx, in_packet=packet, **run_kwargs)

        if self._exit_communication(out_packet=out_packet, ctx=ctx):
            await self._packet_pool.stop_all()
            return

        await self._packet_pool.post(out_packet)

    @property
    def is_listening(self) -> bool:
        return self._is_listening

    async def start_listening(
        self, ctx: RunContext[CtxT] | None = None, **run_kwargs: Any
    ) -> None:
        assert self._packet_pool is not None, "Packet pool must be initialized"

        if self._is_listening:
            return

        self._is_listening = True
        self._packet_pool.register_packet_handler(
            processor_name=self.name,
            handler=self._packet_handler,
            ctx=ctx,
            **run_kwargs,
        )

    async def stop_listening(self) -> None:
        assert self._packet_pool is not None, "Packet pool must be initialized"

        self._is_listening = False
        await self._packet_pool.unregister_packet_handler(self.name)



import time
from typing import Literal, TypeAlias
from uuid import uuid4

from openai.types.chat.chat_completion import ChoiceLogprobs as CompletionChoiceLogprobs
from pydantic import BaseModel, Field, NonNegativeFloat, NonNegativeInt

from .message import AssistantMessage

FinishReason: TypeAlias = Literal[
    "stop", "length", "tool_calls", "content_filter", "function_call"
]


class Usage(BaseModel):
    input_tokens: NonNegativeInt = 0
    output_tokens: NonNegativeInt = 0
    reasoning_tokens: NonNegativeInt | None = None
    cached_tokens: NonNegativeInt | None = None
    cost: NonNegativeFloat | None = None

    def __add__(self, add_usage: "Usage") -> "Usage":
        input_tokens = self.input_tokens + add_usage.input_tokens
        output_tokens = self.output_tokens + add_usage.output_tokens
        if self.reasoning_tokens is not None or add_usage.reasoning_tokens is not None:
            reasoning_tokens = (self.reasoning_tokens or 0) + (
                add_usage.reasoning_tokens or 0
            )
        else:
            reasoning_tokens = None

        if self.cached_tokens is not None or add_usage.cached_tokens is not None:
            cached_tokens = (self.cached_tokens or 0) + (add_usage.cached_tokens or 0)
        else:
            cached_tokens = None

        cost = (
            (self.cost or 0.0) + add_usage.cost
            if (add_usage.cost is not None)
            else None
        )
        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            reasoning_tokens=reasoning_tokens,
            cached_tokens=cached_tokens,
            cost=cost,
        )


class CompletionChoice(BaseModel):
    message: AssistantMessage
    finish_reason: FinishReason | None
    index: int
    logprobs: CompletionChoiceLogprobs | None = None


class CompletionError(BaseModel):
    message: str
    metadata: dict[str, str | None] | None = None
    code: int


class Completion(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    name: str | None = None
    system_fingerprint: str | None = None
    choices: list[CompletionChoice]
    usage: Usage | None = None
    error: CompletionError | None = None

    @property
    def messages(self) -> list[AssistantMessage]:
        return [choice.message for choice in self.choices if choice.message]
